/*

Simplified BSD License
taken from
<http://www.opensource.org/licenses/bsd-license.php> 

Copyright (c) 2008, Saleem N. Bhatti <http://www.cs.st-andrews.ac.uk/~saleem/>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  - Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  - Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

  - Neither the name of 'Saleem N. Bhatti' or 'Saleem Bhatti', nor the
    names of any other contributors may be used to endorse or promote
    products derived from this software without specific prior written
    permission.

DISCLAIMER
----------
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

27 November 2008

*/

/******************************************************************************
Some logging stuff

saleem <http://www.cs.st-andrews.ac.uk/~saleem/>
November 2008
November 2004
November 2001
March 1993
******************************************************************************/

#include "d_Log.h"

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <sys/time.h>
#include <time.h>
#include <pwd.h>
#include <signal.h>
#include <syslog.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

extern "C" {
  int fileno(FILE *);  /* should be in <stdio.h> */
  char *strdup(const char *); /* should be in <string.h> */
  int sigsetmask(const int); /* should be in <signal.h> */
  void vsyslog(int, const char *, va_list); /* should be in <syslog.h> */
}

static void
p_err(const char *format, ...)
{
  va_list args;
 
  va_start(args, format);
 
  if (format == (char *) 0) {
#if defined(DEBUG)
    fprintf(stderr, "%s p_err(): no format string\n", __FILE__);
#endif
    va_end(args);
    return;
  }
 
  vfprintf(stderr, format, args);
  va_end(args);
 
  return;
}


d_Log::d_Log(const char *program, const char *name,
             const unsigned char what, const unsigned char syslog,
             const unsigned char flags, const unsigned long size)
{
  program_ = strdup(program);
  name_ = strdup(program);
  what_ = what;
  syslog_ = syslog;
  flags_ = flags;
  size_ = size;

  log_ = strcmp(name_, "stdout") == 0 ? stdout : (FILE *) 0;

  if (log_ == stdout) {

    /* for stdout we do not need D_LOG_close */
    if (flags_ & D_LOG_close) {
      flags_ &= ~D_LOG_close;
      if (flags_ & D_LOG_alert)
	p_err("d_Log::d_Log(): D_LOG_close removed for stdout\n");
    }
 
    /* warn if logging to stdout and stderr */
    if ((flags_ & D_LOG_stderr) && (flags_ & D_LOG_alert))
      p_err("d_Log::d_Log(): log to stderr (D_LOG_stderr) and stdout\n");
    }
}

d_Log *
d_Log::d_make(const char *program, const char *name,
              const unsigned char what, const unsigned char syslog,
              const unsigned char flags, const unsigned long size)
{

  /* check for program name */
  if (program == (char *) 0) {
    if (flags & D_LOG_alert)
      p_err("d_Log::d_make(): no program name for log\n");
    return (d_Log *) 0;
  }

  /* check for log file name */
  if ((flags & D_LOG_log) && (name == (char *) 0)) {
    if (flags & D_LOG_alert)
      p_err("d_Log::d_make(): no name for log file\n");
    return (d_Log *) 0;
  }

  d_Log *l = new d_Log(program, name, what, flags, syslog, size);
  
  if (l == (d_Log *) 0)
    p_err("d_Log::d_make(): new d_Log() failed!\n");

  return l;
}

d_Log::~d_Log()
{
  if (!(flags_ & D_LOG_close) && log_ != (FILE *) 0)
    d_close();

  free(program_);
  free(name_);
}

void
d_Log::d_open()
{
  if (log_ == stdout)
    return;

  /* check flags */
  if (flags_ & D_LOG_close) {
    if (flags_ & D_LOG_alert)
      p_err("d_Log::d_open(): D_LOG_close specified\n");
    return;
  }

  /* open log */
  if ((log_ = fopen(name_, "a+")) == (FILE *) 0) {
    p_err("d_Log::d_open(): fopen(%s) failed: error %d %s\n",
	  name_, errno, strerror(errno));
    return;
  }

  /* set file to be unbuffered */
  setbuf(log_, (char *) 0);
}

void
d_Log::d_close()
{
  if (log_ == stdout)
    return;

  /* check flags */
  if (flags_ & D_LOG_close) {
    if (flags_ & D_LOG_alert)
      p_err("d_Log::d_close(): D_LOG_close specified\n");
    return;
  }

  /* check to see if file has been opened */
  if (!(flags_ & D_LOG_close) && log_ == (FILE *) 0) {
    if (flags_ & D_LOG_alert)
      p_err("d_Log::d_close(): no open file\n");
    return;
  }

  /* close log */
  if (fclose(log_) != 0) {
    if (flags_ & D_LOG_alert)
      p_err("d_Log::d_close(): fclose(%s) failed\n", name_);
    return;
  }

  log_ = (FILE *) 0;
  
  return;
}

void
d_Log::d_log(const unsigned char what, const char *format, ...)
{
  unsigned flags;
  va_list args;
#define BUFFER_SIZE 1024
    
  char *bp;
  char buffer[BUFFER_SIZE];

  bp = buffer;

  va_start(args, (char *) format);

  /* make message header */
  if ((what_ & what) || (syslog_ & what)) {
 
    /* date and time */
    if (flags_ & (D_LOG_date | D_LOG_time)) {
      struct timeval t;
      struct tm *tp;
      char *tf = "";

      /* both date and time */
      if ((flags_ & D_LOG_date) && (flags_ & D_LOG_time))
	tf = "%H.%M.%S %d-%m-%Y ";

      /* just date */
      else if (flags_ & D_LOG_date)
	tf = "%d-%m-%Y ";

      /* just time */
      else if (flags_ & D_LOG_time)
	tf = "%H.%M.%S ";

      gettimeofday(&t, (struct timezone *) 0);
      tp = localtime(&(t.tv_sec));
      strftime(bp, BUFFER_SIZE, tf, tp);

      bp = buffer + strlen(buffer);
    }

    /* user id info */
    if (flags_ & D_LOG_user) {
      struct passwd *pw = getpwuid(getuid());

      sprintf(bp, "%-8s ", pw->pw_name);
      bp = buffer + strlen(buffer);
    }

    /* process name and id */
    sprintf(bp, "%s %d :", program_, (int) getpid());
    bp = buffer + strlen(buffer);
  }

  /* incorporate header and format info */
  sprintf(bp, "%s\n", format);

  if (what_ & what) {
        
    int mask = sigsetmask(0x7fffffff);

    /* open file */
    if (name_ != (char *) 0 && (flags_ & D_LOG_close)) {

      /* So that d_open() behaves properly */
      flags = flags_;
      flags_ &= ~D_LOG_close;

      d_open();
      if (log_ == (FILE *) 0) {
	flags_ = flags;
	sigsetmask(mask);
	return;
      }
      flags_ = flags;
    }

    /* check size of file */
    if (size_ != 0) {
      struct stat s;

      if (fstat(fileno(log_), &s) != 0) {
 
	/* So that d_close() behaves properly */
	flags = flags_;
	flags_ &= ~D_LOG_close;

	d_close();
	flags_ = flags;
	if (flags_ & D_LOG_alert)
	  p_err("d_Log::log(): fstat() failed\n");

	sigsetmask(mask);
	return;
      }

      if (size_ >= (unsigned long) s.st_size) {
  
	/* So that d_close() behaves properly */
	flags = flags_;
	flags_ &= ~D_LOG_close;

	d_close();
	flags_ = flags;
	if (flags_ & D_LOG_alert)
	  p_err("d_Log::d_log(): size limit exceeded %ld\n", size_);

	sigsetmask(mask);
	return;
      }
    }

    /* do it */
    if (log_ != (FILE *) 0) {

      /* add message to end of file */
      if (fseek(log_, (long) 0, 2) != 0 && (flags_ & D_LOG_alert))
	p_err("d_Log::d_log(): fseek() rewind failed\n");

      vfprintf(log_, buffer, args);
    }

    /* close file */
    if (log_ != (FILE *) 0 && (flags_ & D_LOG_close)) {

      /* So that d_close() behaves properly */
      flags = flags_;
      flags_ &= ~D_LOG_close;

      d_close();
      if (log_ != (FILE *) 0) {
	flags_ = flags;
	sigsetmask(mask);
	return;
      }
      flags_ = flags;
    }

    /* log to stderr */
    if (flags_ & D_LOG_stderr)
      vfprintf(stderr, buffer, args);

    sigsetmask(mask);
  }

  /* write to syslog */
  if (syslog_ & what) {
    int mask = sigsetmask(0x7fffffff);
    vsyslog(LOG_INFO, buffer, args);
    sigsetmask(mask);
  }

    return;
}
