/*

Simplified BSD License
taken from
<http://www.opensource.org/licenses/bsd-license.php> 

Copyright (c) 2008, Saleem N. Bhatti <http://www.cs.st-andrews.ac.uk/~saleem/>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  - Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  - Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

  - Neither the name of 'Saleem N. Bhatti' or 'Saleem Bhatti', nor the
    names of any other contributors may be used to endorse or promote
    products derived from this software without specific prior written
    permission.

DISCLAIMER
----------
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

27 November 2008

*/

/******************************************************************************
Some logging stuff

saleem <http://www.cs.st-andrews.ac.uk/~saleem/>
November 2008
November 2004
November 2001
March 1993
******************************************************************************/

#include "d_log_c.h"

#include <stdlib.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <pwd.h>
#include <signal.h>
#include <syslog.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#if defined(__cplusplus)
extern "C" {
#endif
  int fileno(FILE *);  /* should be in <stdio.h> */
  int sigsetmask(const int); /* should be in <signal.h> */
  static void p_err(const char *format, ...);
  void vsyslog(int, const char *, va_list); /* should be in <syslog.h> */    
#if defined(__cplusplus)
}
#endif

static void
p_err(const char *format, ...)
{
  char *fmt;
  va_list args;
 
  va_start(args, (char *) format);
  fmt = (char *) format;
 
  if (fmt == (char *) 0) {
#if defined(DEBUG)
    fprintf(stderr, "%s p_err(): no format string\n", __FILE__);
#endif
    va_end(args);
    return ;
  }
 
  vfprintf(stderr, fmt, args);
  va_end(args);
 
  return;
}

int
d_log_open(d_Log_t *log)
{
  if (log == (d_Log_t *) 0) {
    if (log->flags & D_LOG_alert)
      p_err("d_log_open(): NULL d_Log_t\n");
    return -1;
  }

  /* check flags */
  if (log->flags & D_LOG_close) {
    if (log->flags & D_LOG_alert)
      p_err("d_log_open(): D_LOG_close specified\n");
    return -1;
  }

  /* check for program name */
  if (log->program == (char *) 0) {
    if (log->flags & D_LOG_alert)
      p_err("d_log_open(): no program name for log\n");
    return -1;
  }

  /* check for log file name */
  if (log->name == (char *) 0) {
    if (log->flags & D_LOG_alert)
      p_err("d_log_open(): no name for log file\n");
    return -1;
  }

  /* open log */
  if ((log->log = fopen(log->name, "a+")) == (FILE *) 0) {
    if (log->flags & D_LOG_alert)
      p_err("d_log_open(): fopen(%s) failed: error %d %s\n",
	    log->name, errno, strerror(errno));
    return -1;
  }

  /* set file to be unbuffered */
  setbuf(log->log, (char *) 0);

  return 0;
}

int
d_log_close(d_Log_t *log)
{
  if (log == (d_Log_t *) 0) {
    if (log->flags & D_LOG_alert)
      p_err("d_log_close(): NULL d_Log_t\n");
    return -1;
  }

  /* check flags */
  if (log->flags & D_LOG_close) {
    if (log->flags & D_LOG_alert)
      p_err("d_log_close(): D_LOG_close specified\n");
    return -1;
  }

  /* check to see if file has been opened */
  if (!(log->flags & D_LOG_close) && log->log == (FILE *) 0) {
    if (log->flags & D_LOG_alert)
      p_err("d_log_close(): no open file\n");
    return -1;
  }

  /* close log */
  if (fclose(log->log) != 0) {
    if (log->flags & D_LOG_alert)
      p_err("d_log_close(): fclose(%s) failed\n", log->name);
    return -1;
  }
  log->log = (FILE *) 0;

  return 0;
}

int
d_log(d_Log_t *log, const unsigned char what, const char *format, ...)
{
  d_Log_t *l;
  unsigned char w, flags;
  char *fmt;
  va_list args;
#define BUFFER_SIZE 1024
  char *bp;
  char buffer[512];

  if (log == (d_Log_t *) 0) {
    if (log->flags & D_LOG_alert)
      p_err("d_log(): NULL d_Log_t\n");
    return -1;
  }

  bp = buffer;

  va_start(args, (char *) format);
  l = log;
  w = (unsigned char) what;
  fmt = (char *) format;

  /* make message header */
  if ((l->what & w) || (l->syslog & w)) {
 
    /* date and time */
    if (l->flags & (D_LOG_date | D_LOG_time)) {
      struct timeval t;
      struct tm *tp;
      char *tf = "";

      /* both date and time */
      if ((l->flags & D_LOG_date) && (l->flags & D_LOG_time))
	tf = "%H.%M.%S %d-%m-%Y ";

      /* just date */
      else if (l->flags & D_LOG_date)
	tf = "%d-%m-%Y ";

      /* just time */
      else if (l->flags & D_LOG_time)
	tf = "%H.%M.%S ";

      gettimeofday(&t, (struct timezone *) 0);
      tp = localtime(&(t.tv_sec));
      strftime(bp, BUFFER_SIZE, tf, tp);

      bp = buffer + strlen(buffer);
    }

    /* user id info */
    if (l->flags & D_LOG_user) {
      struct passwd *pw = getpwuid(getuid());

      sprintf(bp, "%-8s ", pw->pw_name);
      bp = buffer + strlen(buffer);
    }

    /* process name and id */
    sprintf(bp, "%s %d :", l->program, (int) getpid());
    bp = buffer + strlen(buffer);
  }


  /* incorporate header and format info */
  sprintf(bp, "%s\n", format);

  if (l->what & w) {
    int mask = sigsetmask(0x7fffffff);

    /* open file */
    if (l->name != (char *) 0 && (l->flags & D_LOG_close)) {

      /* So that d_log_open() behaves properly */
      flags = l->flags;
      l->flags &= ~D_LOG_close;

      if (d_log_open(l) != 0) {
	l->flags = flags;
	sigsetmask(mask);
	return -1;
      }
      l->flags = flags;
    }

    /* check size of file */
    if (l->size != 0) {
      struct stat s;
 
      if (fstat(fileno(l->log), &s) != 0) {
 
	/* So that d_log_close() behaves properly */
	flags = l->flags;
	l->flags &= ~D_LOG_close;
 
	d_log_close(l);
	l->flags = flags;
	if (l->flags & D_LOG_alert)
	  p_err("d_log_close(): fstat() failed\n");
 
	sigsetmask(mask);
	return -1;
      }
 
      if (l->size >= (unsigned long) s.st_size) {
 
	/* So that d_log_close() behaves properly */
	flags = l->flags;
	l->flags &= ~D_LOG_close;
 
	d_log_close(l);
	l->flags = flags;
	if (l->flags & D_LOG_alert)
	  p_err("d_log_close(): size limit exceeded %ld\n", l->size);
 
	sigsetmask(mask);
	return 0;
      }
    }

    /* do it */
    if (l->log != (FILE *) 0) {

      /* add message to end of file */
      if (fseek(l->log, (long) 0, 2) != 0 && (l->flags & D_LOG_alert))
	if (log->flags & D_LOG_alert)
	  p_err("d_log(): fseek() failed to rewind file\n");
 
      vfprintf(l->log, buffer, args);
      fprintf(l->log, "\n");
    }

    /* close file */
    if (l->log != (FILE *) 0 && (l->flags & D_LOG_close)) {
      /* So that d_log_close() behaves properly */
      flags = l->flags;
      l->flags &= ~D_LOG_close;

      if (d_log_close(l) != 0) {
	l->flags = flags;
	sigsetmask(mask);
	return -1;
      }
      l->flags = flags;
    }

    /* log to stderr */
    if (l->flags & D_LOG_stderr) {
      vfprintf(stderr, buffer, args);
      fprintf(stderr, "\n");
    }

    sigsetmask(mask);
  }

  /* write to syslog */
  if (l->syslog & what) {
    int mask = sigsetmask(0x7fffffff);
    vsyslog(LOG_INFO, buffer, args);
    sigsetmask(mask);
  }

  return 0;
}
